const std = @import("std");

// Music Player:
// - Launch VLC and control it via HTTP interface
// - Play music randomly from a VLC playlist
// - Allow media control (play pause, next, prev)
// - Expose OBS Browser Source to display current track information.
// + Add and remove songs from playlist
// + Allow voting on songs (web interface? chat bot?)
//
// Chat bot:
// - Support Twitch chat
// + Support YouTube
//
// A - means immediately, a + means eventually.

var global_connection: ?std.net.Server.Connection = null;
var global_websocket: ?*std.http.WebSocket = null;

var global_send_buf: ?[]u8 = null;
var global_recv_buf: ?[]align(4) u8 = null;

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();

    var vlc = try VLC.init(allocator, try std.Uri.parse("http://localhost:8080"));
    defer vlc.deinit();

    var tz_file = try std.fs.openFileAbsolute("/etc/localtime", .{});
    defer tz_file.close();
    var tz = try std.tz.Tz.parse(allocator, tz_file.reader());
    defer tz.deinit();

    var server = try std.net.Address.initIp4(.{ 127, 0, 0, 1 }, 8009).listen(.{});
    defer server.deinit();

    // while (true) : (std.time.sleep(1_000_000_000)) { // sleep 500ms
    //     try updateTime(tz);
    //     try updateStatus(allocator, &vlc);
    //     // scroll += 1;
    // }

    defer if (global_connection) |conn| conn.stream.close();
    defer if (global_send_buf) |buf| allocator.free(buf);
    defer if (global_recv_buf) |buf| allocator.free(buf);

    var delay: u8 = 0;
    while (true) {
        var pollfds = [_]std.posix.pollfd{.{ .fd = server.stream.handle, .events = std.posix.POLL.IN, .revents = 0 }};
        const num = try std.posix.poll(&pollfds, 100);
        std.debug.print("num pollfds {d}\nrevents {d}\n", .{ num, pollfds[0].revents });
        if (pollfds[0].revents != 0) {
            const read_buf = try allocator.alloc(u8, 1024 * 32);
            defer allocator.free(read_buf);

            const connection = try server.accept();
            errdefer connection.stream.close();

            var http = std.http.Server.init(connection, read_buf);
            var request = try http.receiveHead();
            std.debug.print("Target: {s}\n", .{request.head.target});

            var websocket: std.http.WebSocket = undefined;

            global_send_buf = try allocator.alloc(u8, 1024 * 32);
            global_recv_buf = try allocator.alignedAlloc(u8, 4, 1024 * 32);

            const is_websocket = try websocket.init(&request, global_send_buf.?, global_recv_buf.?);
            if (is_websocket) {
                try websocket.response.flush();
                std.debug.print("is a websocket now\n", .{});
                global_connection = connection;
                global_websocket = &websocket;
                // const msg = try websocket.readSmallMessage();
                // std.debug.print("{any}\n{s}\n", .{ msg, msg.data });

                // try websocket.writeMessage("dude i cant believe you just ate that", .text);
                // // try websocket.writeMessage("", .ping);
                // std.time.sleep(5_000_000_000);
                // try websocket.writeMessage("dude why did i just eat that now", .text);
                // std.debug.print("sent second one\n", .{});
            } else {
                defer connection.stream.close();
                switch (request.head.method) {
                    .GET, .HEAD => {
                        if (std.mem.eql(u8, request.head.target, "/gone")) {
                            try request.respond(@embedFile("web/gone.html"), .{});
                        } else if (std.mem.eql(u8, request.head.target, "/statusline")) {
                            try request.respond(@embedFile("web/statusline.html"), .{});
                        } else if (std.mem.eql(u8, request.head.target, "/gone.js")) {
                            try request.respond(@embedFile("web/gone.js"), .{});
                        } else if (std.mem.eql(u8, request.head.target, "/statusline.js")) {
                            try request.respond(@embedFile("web/statusline.js"), .{});
                        } else if (std.mem.eql(u8, request.head.target, "/style.css")) {
                            try request.respond(@embedFile("web/style.css"), .{});
                        } else {
                            try request.respond("", .{ .status = .not_found });
                        }
                    },
                    else => try request.respond("", .{ .status = .not_found }),
                }
            }
        }

        delay += 1;
        if (delay == 10) {
            delay = 0;
            try updateStatus(allocator, &vlc);
        }
    }
}

const base64_encoder = std.base64.standard.Encoder;

// TODO make the URL something short like jeevio.xyz/streamboy
const topic = "Something Fun For Everyone With Streamboy!!!!!!!! git.jeevio.xyz/jeeves/streamboy";

// var scroll: usize = 0;

fn updateStatus(allocator: std.mem.Allocator, vlc: *VLC) !void {
    // var stream_info_file = try std.fs.createFileAbsolute("/tmp/streaminfo", .{ .truncate = true });
    // defer stream_info_file.close();
    if (global_websocket) |ws| {
        var song_info = try vlc.getSongInfo();
        defer song_info.deinit();

        const string = try std.fmt.allocPrint(allocator, "{s}    |    ♪ {s} - {s}    |    ", .{
            topic,
            song_info.title orelse "Unknown Title",
            song_info.artist orelse "Unknown Artist",
        });
        defer allocator.free(string);

        try ws.writeMessage(string, .text);
    }

    // if (scroll > string.len) scroll = 0;
    // if (scroll == 0) try stream_info_file.writeAll(string) else {
    //     for (string[scroll..]) |char| try stream_info_file.writer().writeByte(char);
    //     for (string[0..scroll]) |char| try stream_info_file.writer().writeByte(char);
    // }
}

pub const VLC = struct {
    allocator: std.mem.Allocator,
    http: std.http.Client,
    base_uri: std.Uri,
    authorization: []u8,

    pub fn init(allocator: std.mem.Allocator, base_uri: std.Uri) !VLC {
        const userpass = try std.fmt.allocPrint(allocator, ":{s}", .{"1234"});
        defer allocator.free(userpass);

        const base64_userpass = try allocator.alloc(u8, base64_encoder.calcSize(userpass.len));
        defer allocator.free(base64_userpass);

        const authorization = try std.fmt.allocPrint(allocator, "Basic {s}", .{base64_encoder.encode(base64_userpass, userpass)});
        errdefer allocator.free(authorization);

        // var vlc = std.process.Child.init(&[_][]const u8{
        //     "vlc",
        //     "--intf",
        //     "http",
        //     "--http-host",
        //     "localhost",
        //     "--http-password",
        //     "1234",
        // }, allocator);
        // try vlc.spawn();

        return .{
            .allocator = allocator,
            .http = std.http.Client{ .allocator = allocator },
            .base_uri = base_uri,
            .authorization = authorization,
        };
    }

    pub fn deinit(self: *VLC) void {
        self.http.deinit();
        self.allocator.free(self.authorization);
        // try vlc.kill();
    }

    fn request(self: *VLC, uri: std.Uri) !Response {
        var combined_uri = self.base_uri;
        combined_uri.path = uri.path;
        combined_uri.query = uri.query;
        combined_uri.fragment = uri.fragment;

        var response = std.ArrayList(u8).init(self.allocator);
        defer response.deinit();

        const result = try self.http.fetch(.{
            .location = .{ .uri = combined_uri },
            .headers = .{ .authorization = .{ .override = self.authorization } },
            .response_storage = .{ .dynamic = &response },
        });

        // std.debug.print("{any}\n{s}\n", .{ result, response.items });

        if (result.status != .ok) return error.HttpRequestFailed;

        const buffer = try response.toOwnedSlice();
        return .{
            .buffer = buffer,
            .document = try xml.parse(self.allocator, buffer),
        };
    }

    const Response = struct {
        buffer: []u8,
        document: xml.Document,

        pub fn deinit(self: *Response, allocator: std.mem.Allocator) void {
            self.document.deinit();
            allocator.free(self.buffer);
        }
    };

    pub const SongInfo = struct {
        allocator: std.mem.Allocator,
        title: ?[]const u8,
        album: ?[]const u8,
        artist: ?[]const u8,

        // TODO move allocator into struct
        pub fn deinit(self: *SongInfo) void {
            if (self.title) |b| self.allocator.free(b);
            if (self.album) |b| self.allocator.free(b);
            if (self.artist) |b| self.allocator.free(b);
        }
    };

    pub fn getSongInfo(self: *VLC) !SongInfo {
        var response = try self.request(.{ .scheme = "http", .path = .{ .percent_encoded = "/requests/status.xml" } });
        defer response.deinit(self.allocator);

        // std.debug.print("{s}\n", .{response.buffer});

        var title: ?[]const u8 = null;
        var album: ?[]const u8 = null;
        var artist: ?[]const u8 = null;

        if (response.document.root.findChildByTag("information")) |information| {
            var categories_it = information.findChildrenByTag("category");
            while (categories_it.next()) |category| {
                if (std.mem.eql(u8, category.getAttribute("name").?, "meta")) {
                    var info_it = category.findChildrenByTag("info");
                    while (info_it.next()) |info| {
                        const info_name = info.getAttribute("name").?;
                        if (std.mem.eql(u8, info_name, "title"))
                            title = try processHtmlString(self, info.children[0].char_data)
                        else if (std.mem.eql(u8, info_name, "album"))
                            album = try processHtmlString(self, info.children[0].char_data)
                        else if (std.mem.eql(u8, info_name, "artist"))
                            artist = try processHtmlString(self, info.children[0].char_data);
                    }
                }
            }
        }

        return .{
            .allocator = self.allocator,
            .title = title,
            .album = album,
            .artist = artist,
        };
    }

    // TODO clean up
    fn processHtmlString(self: *VLC, string: []const u8) ![]const u8 {
        var new: []u8 = try self.allocator.dupe(u8, string);
        errdefer self.allocator.free(new);
        while (true) {
            if (std.mem.indexOf(u8, new, "&#")) |amp| {
                if (std.mem.indexOfScalarPos(u8, new, amp, ';')) |semi| {
                    const int = try std.fmt.parseInt(u8, new[amp + 2 .. semi], 10);
                    const nnew = try self.allocator.alloc(u8, std.mem.replacementSize(u8, new, new[amp .. semi + 1], &[1]u8{int}));
                    _ = std.mem.replace(u8, new, new[amp .. semi + 1], &[1]u8{int}, nnew);
                    self.allocator.free(new);
                    new = nnew;
                }
            } else break;
        }
        // std.debug.print("{s}\n", .{new});
        return new;
    }

    const xml = @import("./xml.zig");
};

fn updateTime(tz: std.Tz) !void {
    const original_timestamp = std.time.timestamp();

    var timetype: *std.tz.Timetype = undefined;
    for (tz.transitions, 0..) |trans, i|
        if (trans.ts >= original_timestamp) {
            timetype = tz.transitions[i - 1].timetype;
            break;
        };

    const timestamp = std.time.timestamp() + timetype.offset + std.time.s_per_day;

    const epoch_seconds = std.time.epoch.EpochSeconds{ .secs = @intCast(timestamp) };
    const day_seconds = epoch_seconds.getDaySeconds();
    const epoch_day = epoch_seconds.getEpochDay();
    const year_day = epoch_day.calculateYearDay();
    const month_day = year_day.calculateMonthDay();

    const ampm = getAmPm(day_seconds);

    var file = try std.fs.createFileAbsolute("/tmp/time", .{ .truncate = true });
    defer file.close();

    try file.writer().print("{s}, {s} {d} {d}, {d:0>2}:{d:0>2}:{d:0>2} {s}", .{
        getDayOfWeekName(getDayOfWeek(epoch_day)),
        getMonthName(month_day.month),
        month_day.day_index,
        year_day.year,
        ampm.hour,
        day_seconds.getMinutesIntoHour(),
        day_seconds.getSecondsIntoMinute(),
        if (ampm.is_pm) "PM" else "AM",
    });
}

fn getMonthName(month: std.time.epoch.Month) []const u8 {
    return switch (month) {
        .jan => "January",
        .feb => "February",
        .mar => "March",
        .apr => "April",
        .may => "May",
        .jun => "June",
        .jul => "July",
        .aug => "August",
        .sep => "September",
        .oct => "October",
        .nov => "November",
        .dec => "December",
    };
}

// Jan 1 1970 was a Thursday, so we make that 1
const DayOfWeek = enum(u3) {
    wed = 0,
    thu,
    fri,
    sat,
    sun,
    mon,
    tue,
};

fn getDayOfWeek(epoch_day: std.time.epoch.EpochDay) DayOfWeek {
    return @enumFromInt(@mod(epoch_day.day, 7));
}

fn getDayOfWeekName(day_of_week: DayOfWeek) []const u8 {
    return switch (day_of_week) {
        .sun => "Sunday",
        .mon => "Monday",
        .tue => "Tuesday",
        .wed => "Wednesday",
        .thu => "Thursday",
        .fri => "Friday",
        .sat => "Saturday",
    };
}

const AmPm = struct {
    hour: u4,
    is_pm: bool,
};

fn getAmPm(day_seconds: std.time.epoch.DaySeconds) AmPm {
    const hour = day_seconds.getHoursIntoDay();
    return .{
        .hour = if (hour < 13) @intCast(hour) else @intCast(hour - 12),
        .is_pm = hour != 0 and hour > 11,
    };
}
